'''
Created on Sep 8, 2021

@author: don_bacon
'''

VERSION_TUPLE = (0, 1, 2)
__version__ = ".".join(map(str, VERSION_TUPLE))
__title__ = 'common'
__description__ = 'Python Common library'
__author__ = 'Donald Bacon'
__author_email__ = 'dwbzen@gmail.com'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2022 Donald Bacon'